/**
 * HM acquire - software for reading out HAMEG digital storage
 * oscilloscopes.
 *         
 * Copyright (C) 2007 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@gmx.de
 * http://formica.nusseis.de/HMacquire/
 */


/*
 * This routines work on pure channel data (double *)
 * accompained by scaling information 
 *
 * Uses FFTW 3 interface by default. May use FFTW 2 interface if compiled with
 * -DFFTW2
 */
#ifndef DATA_API_H
#define DATA_API_H

#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef FFTW2
#include <rfftw.h>
#else
#include <fftw3.h>
#endif

#define DATA_MARK "--- hmacquire channel data ---"

/**
 * print xy-data to file
 * output can be read by read_xy().
 *
 * f	outpuf file
 * xb	x data to print
 * yb	y data to print
 * xdiv voltage per devision
 * ydiv voltage per devision
 */
void print_xy( FILE * f, const double * xb, const double * yb, 
			 double xdiv, double ydiv );

/**
 * read xy data from file f
 *
 * seeks for DATA_MARK and then reads output from print_xy 
 * data      is stored to xb, yb
 * voltscale is stored to xdiv, ydiv
 * returns 1 on success, 0 on error
 */
int read_xy( FILE * f, double * xb, double * yb, double * xdiv, double * ydiv );

/**
 * print data for cahannel to file f
 * output can be read by read_chan().
 *
 * f	outpuf file
 * b	data to print
 * tdiv time per division, i.e. per 200 data points
 * ydiv voltage per devision
 */
void print_chan( FILE * f, const double * b, double tdiv, double ydiv );


/**
 * read data from file f
 *
 * seeks for DATA_MARK and then reads output from print_chan 
 * data      is stored to b
 * timscale  is stored to tdiv
 * voltscale is stored to ydiv
 * returns 1 on success, 0 on error
 */
int read_chan( FILE * f, double * b, double * tdiv, double * ydiv );


/**
 * Add two channels.
 *
 * Destination may be one of the sources.
 *
 * d	destination
 * s1	source 1
 * s2	source2
 */
void chan_add( double * d, const double * s1, const double * s2  );


/**
 * Calculate spectrum for channel
 * The values of s are given in dBV
 *
 * c  channel data, array length 2048
 * s  spectrum, array length 1025
 */
void chan_spec( const double * c, double * s );


/**
 * print spectrum to file f
 *
 * f	outpuf file
 * b	data to print
 * tdiv time per division, i.e. per 200 data points
 * ydiv voltage per devision
 */
void print_spec( FILE * f, const double * b, double tdiv, double ydiv );


/**
 * Calculate correlation function of f and g.
 *
 * d	destination
 * s1	source 1
 * s2	source 2
 */
void chan_corr( double * d, const double * s1, const double * s2  );


/**
 * print correlation to file f
 *
 * f	outpuf file
 * b	data to print
 * tdiv time per division, i.e. per 200 data points
 */
void print_corr( FILE * f, const double * b, double tdiv, double ydiv );

#endif
