/**
 * HM acquire - software for reading out HAMEG digital storage
 * oscilloscopes.
 *         
 * Copyright (C) 2007, 2011, 2014, 2018 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@opppf.de
 * http://opppf.de/HMacquire/
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include "osc_api.h"
#include "data_api.h"

/**********************************************************************
				default values
 **********************************************************************/


#define GNUPLOT "gnuplot"
#define LPR "|lpr"
#define SERIALPORT "/dev/ttyS1"

#define VERSION "1.8 (2018-05-04)"

/**********************************************************************
				  data types
 **********************************************************************/


/* flag to indicate output file type: channel, xy, spectrum, or correaltion  */
#define TCHAN 1
#define TSPEC 2
#define TCORR 4
#define TXY 8


/* length of user input buffers */
#define INTBUFLEN 256


/** hold data for job */
typedef struct 
{
  char source;		/* 1, 2, Both, Add, XY, Osc, File */
  int ch;		/* boolean: print channel */
  int sp;		/* boolean: print spectrum */
  int co;		/* boolean: print correlation */
  char format;		/* X11, Eps, Ps, pNg, Gnuplot, Dat */
  char * out;		/* basename for outfiles */
  char * in;		/* infile */
  char * serial;	/* name of serial port */
  int interactive;	/* booelan: interacitve mode */
  double probes[2];	/* factor of voltage proges */
} tjob;


/**********************************************************************
			       helper functions
 **********************************************************************/

/**
 * determine which channel is demanded
 * returns job.source, but resloves o (scope state) and f (file)
 * return: 1, 2, a (add), b (both), y (XY)
 */
char resolve_source( const tjob * job, const OSTATE * state )
{
  /* active channel */
  if ( index( "of", job->source ) )
  {
    if ( state->add ) return 'a';
    if ( state->XY ) return 'y';
    if ( state->Yon[0] && state->Yon[1] ) return 'b';
    if ( state->Yon[0] ) return '1';
    if ( state->Yon[1] ) return '2';
  }
  /* fixed channels */
  return job->source;
}

/**
 * generate output filename
 * Filename: basename.type.ending
 * type is TCHAN or TSPEC or TCORR 
 */
char * generate_outfile_name( const tjob * job, int type  )
{
  char * filename;
  char * typename;
  char * ending;

  filename = malloc( ( strlen( job->out ) + 10 ) * sizeof(char) ); 
  switch ( job->format )
  {
    case 'e': ending = "eps"; break;
    case 'p': ending = "ps"; break;
    case 'g': ending = "plt"; break;
    case 'd': ending = "dat"; break;
    case 'n': ending = "png"; break;
    default: 
      fputs( "Unknown outfile in generate_outfile_name. It's a bug!\n", 
             stderr );
      exit( 1 );
  }
  switch ( type )
  {
    case TCHAN: typename = "chan"; break;
    case TSPEC: typename = "spec"; break;
    case TCORR: typename = "corr"; break;
    case TXY: typename = "xy"; break;
    default:
      fputs( "Unknown type in generate_outfile_name. It's a bug!\n", stderr );
      exit( 1 );
  }
  sprintf( filename, "%s.%s.%s", job->out, typename, ending );
  return filename;
}

/**
 * open file or stdout
 * Filename: basename.type.ending
 * type is TCHAN or TSPEC or TCORR or TX
 */
FILE * open_outfile( const tjob * job, int type  )
{
  FILE * o;
  char * filename;
  
  if ( job->out )
  {
    filename = generate_outfile_name( job, type );
    o = fopen( filename, "w" );
    /* error handling */
    if ( ! o )
    {
      fprintf( stderr, "Can't open file '%s'.\n", filename );
      exit( 1 );
    }
    /* clean up */
    free( filename );
    return o;
  }
  else
    return stdout;
}


/**
 * close outfile
 */
void close_outfile( const tjob * job, FILE * o )
{
  if ( job->out )
    fclose( o );
}


/**
 * open gnuplot FILE
 */
FILE * open_gnuplot()
{
  FILE * gp;

  gp = popen( GNUPLOT, "w" );
  if ( gp == 0 )
  {
    fprintf( stderr, "Can't open '%s'.\n", GNUPLOT );
    exit( 1 );
  }
  return gp;
}


/**
 * close gnuplot 
 */
void close_gnuplot( FILE * gp )
{
  pclose( gp );
}


/**
 * Generate neat scale
 *
 * generates s and f so that
 * s = x * f
 *
 * s	is in [0.1 .. 1000[
 * f	is a multiple of 1000
 * p	is a string holding the prefix char. 
 *	p must have at least lenght of 2 chars!
 */
void neat_scale( double x, double * s, double * f, char * p )
{
  *s = log( x ) / log( 1000 );
  *f = floor( *s );
  switch ( (int)*f )
  {
    case -8: *p = 'y'; break;
    case -7: *p = 'z'; break;
    case -6: *p = 'a'; break;
    case -5: *p = 'f'; break;
    case -4: *p = 'p'; break;
    case -3: *p = 'n'; break;
    case -2: *p = 181; break; /* µ in iso-latin-1 */
    case -1: *p = 'm'; break;
    case  0: *p = 0; break;   /* empty string */
    case  1: *p = 'k'; break;
    case  2: *p = 'M'; break;
    case  3: *p = 'G'; break;
    case  4: *p = 'T'; break;
    case  5: *p = 'P'; break;
    case  6: *p = 'E'; break;
    case  7: *p = 'Z'; break;
    case  8: *p = 'Y'; break;
    default: *p = '?'; 
  }
  p[1] = 0; /* terminate string */
  *s = pow( 1000, *s - *f );
  *f = pow( 1000, - *f );
}


/**
 * print gnuplot header
 * 
 * type is TCHAN or TSPEC or TCORR or TXY
 */
void gnuplot_header( const tjob * job, const OSTATE * state, int type,
                     FILE * gp )
{
  char * outfile;

  fputs( "#!/usr/bin/gnuplot -persist\n", gp );  
  fputs( "# GNUplot script made by hmacquire\n", gp );  
  fputs( "reset\n", gp ); 
  fputs( "\n", gp );
  /* select terminal */
  switch ( job->format )
  {
    case 'e':
    case 'g':
      fputs( "set terminal postscript eps colour dl 3 lw 3 28\n", gp );
      /* Warum geht das nicht? Eps-Datei wird angelegt, Größe 0.
       * Aufruf der erzeugten Gnuplot-Datei geht aber.
      fputs( "set terminal epscairo font ',16' linewidth 3\n", gp );
      */
      fputs( "set size ratio 4.0/5.0\n", gp ); 
      break;
    case 'p':
      fputs( "set terminal postscript landscape dl 3 lw 3 24\n", gp );
      break;
    case 'n':
      fputs( "set terminal pngcairo fontscale 1.2 linewidth 2\n", gp );
      break;
    case 'x':
      fputs( "set mouse\n", gp );
      switch ( type )
      {
	case TCHAN: 
	case TXY: 
	  fputs( "set terminal x11 0 title 'Monitor HM'\n", gp );
	  break;
	case TSPEC:
	  fputs( "set terminal x11 1 title 'Monitor HM spectrum'\n", gp );
	  break;
	case TCORR:
	  fputs( "set terminal x11 2 title 'Monitor HM correlation'\n", gp );
	  break;
      }
      break;
  }
  /* select output file name */
  if ( job->out )
  {
    if ( job->out[0] == '|' )
    {
      /* print to other program, e.g. lpr */
      fprintf( gp, "set output '%s'\n", job->out );
    }
    else
    {
      /* generate outfile from basename */
      outfile = generate_outfile_name( job, type );
      if ( 'g' == job->format )
	strcpy( rindex( outfile, '.' ), ".eps" ); /* write .eps in plt file */
      fprintf( gp, "set output '%s'\n", outfile );
      free( outfile );
    }
  }
  /* common header */
  fputs( "\n", gp );
  fputs( "set encoding iso_8859_1\n", gp );
  fputs( "set decimalsign ','\n", gp );
  fputs( "set arrow from graph 0.5, 0 rto graph 0, 1 nohead lt 0 lw 2\n", gp );
  fputs( "set arrow from graph 0, 0.5 rto graph 1, 0 nohead lt 0 lw 2\n", gp );
  fputs( "set arrow from graph 0, 0.25 rto graph 1, 0 nohead lt 0 lw 2\n", gp );
  fputs( "set arrow from graph 0, 0.75 rto graph 1, 0 nohead lt 0 lw 2\n", gp );
  fputs( "set grid xtics mxtics ytics\n", gp );
  fputs( "set tics scale 1.0, 0.5, 0.5, 0.0\n", gp );
  fputs( "set style data lines\n", gp );
  fputs( "set key inside right top horizontal reverse samplen 2\n", gp );
  fputs( "\n", gp );
}


/**********************************************************************
				data delivery
 **********************************************************************/


/** 
 * deliver correlation to datafile
 */
void job_dlvr_corr_data( const tjob * job, const OSTATE * state, 
                       const double * corr, FILE * out )
{
  double ydiv;
  switch ( resolve_source( job, state ) )
  {
    case '1': 
      fputs( "# auto_correlation( CH1 )\n", out );
      ydiv = state->Yv[0] * state->Yv[0];
      break;
    case '2':
      fputs( "# auto_correlation( CH2 )\n", out );
      ydiv = state->Yv[1] * state->Yv[1];
      break;
    case 'b': 
      fputs( "# cross_correlation( CH1, CH2 )\n", out );
      ydiv = state->Yv[0] * state->Yv[1];
      break;
    case 'a': 
      fprintf( out, "# auto_correlation( %cCH 1 %c CH 2 )\n", 
              (state->Ys[0] == 1)?'+':'-', (state->Ys[1] == 1)?'+':'-' );
      ydiv = state->Yv[0] * state->Yv[0];
      break;
    default:
      fputs( "Error in job_dlvr_corr_data. This is a bug!\n", stderr );
      exit( 1 );
  }
  print_corr( out, corr, state->Ts, ydiv );
}


/** 
 * deliver correlation to gnuplot
 */
void job_dlvr_corr_gp( const tjob * job, const OSTATE * state, 
                       const double * corr, FILE * gp )
{
  double tdiv, texp;
  double ydiv;
  char tpre[2];
  
  /* prepare scales */
  neat_scale( state->Ts, &tdiv, &texp, tpre );
  /* write gnuplot file */
  gnuplot_header( job, state, TCORR, gp );
  /* time axis */
  fprintf( gp, "set xrange [%.1f:%.1f]\n", -10*tdiv, 10*tdiv );
  fputs( "set mxtics 2\n", gp );
  fprintf( gp, "set xtics %.1f\n",  2*tdiv );
  fprintf( gp, "set xtics add ('%ss' %f 2)\n", tpre, 8*tdiv );
  /* voltage axis */
  fputs( "set ytics autofreq\n", gp );
  fputs( "set yrange [*:*]\n", gp );
  switch ( resolve_source( job, state ) )
  {
    case '1': 
      fputs( "set title 'auto-correlation( CH1 )'\n", gp );
      ydiv = state->Yv[0] * state->Yv[0];
      break;
    case '2':
      fputs( "set title 'auto-correlation( CH2 )'\n", gp );
      ydiv = state->Yv[1] * state->Yv[1];
      break;
    case 'b': 
      fputs( "set title 'cross-correlation( CH1, CH2 )'\n", gp );
      ydiv = state->Yv[0] * state->Yv[1];
      break;
    case 'a': 
      fprintf( gp, "set title 'auto-correlation( %cCH1 %c CH2 )'\n", 
              (state->Ys[0] == 1)?'+':'-', (state->Ys[1] == 1)?'+':'-' );
      ydiv = state->Yv[0] * state->Yv[0];
      break;
  }
  /* plot correlation */
  fputs( "\n", gp );  
  fprintf( gp, "plot '-' using ($1*%e):2 notitle\n", texp );  
  print_corr( gp, corr, state->Ts, ydiv );
  fputs( "e\n", gp );
}


/** 
 * deliver spectrum to datafile
 */
void job_dlvr_spec_data( const tjob * job, const OSTATE * state, 
                       const double * spec[2], FILE * out )
{
  /* print chan 1 */
  if ( index( "1b", resolve_source( job, state ) ) )
  {
    fputs( "# CH 1\n", out );
    print_spec( out, spec[0], state->Ts, state->Yv[0] );
  }
  if ( 'a' == resolve_source( job, state ) )
  {
    fprintf( out, "# %cCH 1 %c CH 2\n", 
	     (state->Ys[0] == 1)?'+':'-', (state->Ys[1] == 1)?'+':'-' );
    print_spec( out, spec[0], state->Ts, state->Yv[0] );
  }
  /* print chan 2 */
  if ( 'b' == resolve_source( job, state ) )
    fputs( "\n\n", out );
  if ( index( "2b", resolve_source( job, state ) ) )
  {
    fputs( "# CH 2\n", out );
    print_spec( out, spec[1], state->Ts, state->Yv[1] );
  }
}


/** 
 * deliver spectrum to gnuplot
 */
void job_dlvr_spec_gp( const tjob * job, const OSTATE * state, 
                       const double * spec[2], FILE * gp )
{
  double fN, L1max, L2max, Lmax, Lmin;
  double fdiv, fexp;
  char fpre[2];

  /* prepare scaling */
  fN = 100.0 / state->Ts;
  neat_scale( fN, &fdiv, &fexp, fpre );
  L1max = 4 * state->Yv[0] * 4/M_PI; /* maximum ampl. in most Fourier sums */
  L1max = 20 * log( L1max ) / log( 10 );
  L2max = 4 * state->Yv[1] * 4/M_PI;
  L2max = 20 * log( L2max ) / log( 10 );
  switch ( resolve_source( job, state ) )
  {
    case 'a':
    case '1': 
      Lmax = L1max;
      Lmin = L1max - 60;
      break;
    case '2': 
      Lmax = L2max;
      Lmin = L2max - 60;
      break;
    case 'b': 
      Lmax = ( L1max > L2max )?L1max:L2max;
      Lmin = (( L1max < L2max )?L1max:L2max) - 60;
      break;
  }
  /* write gnuplot file */
  gnuplot_header( job, state, TSPEC, gp );
  /* frequency axis */
  fprintf( gp, "set xrange [0:%.1f]\n", fdiv );
  fputs( "set mxtics 2\n", gp );
  fprintf( gp, "set xtics add ('%sHz' %f 2)\n", fpre, 0.875*fdiv );
  /* Power axis */
  fputs( "set ytics 10\n", gp );
  fputs( "set mytics 2\n", gp );
  fprintf( gp, "set ytics add ('dBV' %f 3)\n", 10*floor(Lmax/10)-5 );
  fprintf( gp, "set yrange [%.0f:%.0f]\n", Lmin, Lmax );
  fputs( "plot ", gp );  
  /* plot chan 1 */
  if ( index( "1ba", resolve_source( job, state ) ) )
    fprintf( gp, "'-' using ($1*%e):2", fexp );  
  switch ( resolve_source( job, state ) )
  {
    case '1': fputs( " notitle", gp ); break;
    case 'b': fputs( " title 'CH 1'", gp ); break;
    case 'a': fputs( " notitle", gp ); break;
  }
  /* plot chan 2 */
  if ( 'b' == resolve_source( job, state ) )
    fputs( ",\\\n     ", gp );
  if ( index( "2b", resolve_source( job, state ) ) )
    fprintf( gp, "'-' using ($1*%e):2", fexp );  
  switch ( resolve_source( job, state ) )
  {
    case '2': fputs( " notitle", gp ); break;
    case 'b': fputs( " title 'CH 2'", gp ); break;
  }
  fputs( "\n", gp );  
  /* print data */
  if ( index( "1ba", resolve_source( job, state ) ) )
  {
    print_spec( gp, spec[0], state->Ts, state->Yv[0] );
    fputs( "e\n", gp );
  }
  if ( index( "2b", resolve_source( job, state ) ) )
  {
    print_spec( gp, spec[1], state->Ts, state->Yv[1] );
    fputs( "e\n", gp );
  } 
}


/** 
 * deliver channel data to datafile
 */
void job_dlvr_chan_data( const tjob * job, const OSTATE * state, 
                       const double * chan[2], FILE * out )
{
  /* print chan 1 */
  if ( index( "1b", resolve_source( job, state ) ) )
  {
    fputs( "# CH 1\n", out );
    print_chan( out, chan[0], state->Ts, state->Yv[0] );
  }
  if ( 'a' == resolve_source( job, state ) )
  {
    fprintf( out, "# %cCH 1 %c CH 2\n", 
	     (state->Ys[0] == 1)?'+':'-', (state->Ys[1] == 1)?'+':'-' );
    print_chan( out, chan[0], state->Ts, state->Yv[0] );
  }
  /* print chan 2 */
  if ( 'b' == resolve_source( job, state ) )
    fputs( "\n\n", out );
  if ( index( "2b", resolve_source( job, state ) ) )
  {
    fputs( "# CH 2\n", out );
    print_chan( out, chan[1], state->Ts, state->Yv[1] );
  }
}


/** 
 * deliver channel data to gnuplot
 */
void job_dlvr_chan_gp( const tjob * job, const OSTATE * state, 
                       const double * chan[2], FILE * gp )
{
  double tdiv, texp, y1div, y1exp, y2div, y2exp;
  char tpre[2], y1pre[2], y2pre[2];
  
  /* prepare scales */
  neat_scale( 10 * state->Ts, &tdiv, &texp, tpre );
  neat_scale( state->Yv[0], &y1div, &y1exp, y1pre );
  neat_scale( state->Yv[1], &y2div, &y2exp, y2pre );
  tdiv /= 10;
  /* write gnuplot file */
  gnuplot_header( job, state, TCHAN, gp );
  /* time axis */
  fprintf( gp, "set xrange [%.1f:%.1f]\n", 0.0, 10*tdiv );
  fputs( "set mxtics 2\n", gp );
  fprintf( gp, "set xtics %.1f\n",  2*tdiv );
  fprintf( gp, "set xtics add ('%ss' %.1f 2)\n", tpre, 9*tdiv );
  /* voltage axis */
  switch ( resolve_source( job, state ) )
  {
    case 'b':
      fprintf( gp, "set y2tics %.0f textcolor lt 2\n", y2div );
      fprintf( gp, "set y2tics add (' %sV' %f 3)\n", y2pre, 3.5*y2div );
      fprintf( gp, "set y2range [%.0f:%.0f]\n", -4*y2div, 4*y2div );
      fprintf( gp, "set ytics textcolor lt 1\n" );
      /* continue with 1,; no break; */
    case '1':
    case 'a':
      fprintf( gp, "set ytics %.0f\n", y1div );
      fprintf( gp, "set ytics add ('%sV' %f 3)\n", y1pre, 3.5*y1div );
      fprintf( gp, "set yrange [%.0f:%.0f]\n",  -4*y1div, 4*y1div );
      break;
    case '2':
      fprintf( gp, "set ytics %.0f\n", y2div );
      fprintf( gp, "set ytics add ('%sV' %f 3)\n", y2pre, 3.5*y2div );
      fprintf( gp, "set yrange [%.0f:%.0f]\n", -4*y2div, 4*y2div );
      break;
  }
  fputs( "\n", gp );  
  fputs( "plot ", gp );  
  /* plot chan 1 */
  if ( index( "1ba", resolve_source( job, state ) ) )
    fprintf( gp, "'-' using ($1*%e):($2*%e)", texp, y1exp );  
  switch ( resolve_source( job, state ) )
  {
    case '1':
    case 'b': 
	fprintf( gp, " title 'CH1 %cC%s'", 
		     state->Yc[0], 
	             (state->Ys[0] == 1)?"":" INV" );
	break;
    case 'a': 
	fprintf( gp, " title '(%c CH1 %cC %c CH2 %cC)'\n", 
                     (state->Ys[0] == 1)?' ':'-', state->Yc[0], 
	             (state->Ys[1] == 1)?'+':'-', state->Yc[0]);
	break;
  }
  /* plot chan 2 */
  if ( 'b' == resolve_source( job, state ) )
    fputs( ",\\\n     ", gp );
  if ( index( "2b", resolve_source( job, state ) ) )
    fprintf( gp, "'-' using ($1*%e):($2*%e)", texp, y2exp );  
  if ( 'b' == resolve_source( job, state ) )
    fputs( " axis x1y2", gp ); 
  if ( index( "2b", resolve_source( job, state ) ) )
    fprintf( gp, " title 'CH2 %cC%s'", state->Yc[1], 
				       (state->Ys[1] == 1)?"":" INV" );
  fputs( "\n", gp );  
  /* print data */
  if ( index( "1ba", resolve_source( job, state ) ) )
  {
    print_chan( gp, chan[0], state->Ts, state->Yv[0] );
    fputs( "e\n", gp );
  }
  if ( index( "2b", resolve_source( job, state ) ) )
  {
    print_chan( gp, chan[1], state->Ts, state->Yv[1] );
    fputs( "e\n", gp );
  }
}

/** 
 * deliver xy data to datafile
 */
void job_dlvr_xy_data( const OSTATE * state, 
                       const double * chan[2], FILE * out )
{
    fputs( "# XY\n", out );
    print_xy( out, chan[1], chan[0], state->Yv[1], state->Yv[0] );
}

/** 
 * deliver xy data to gnuplot
 */
void job_dlvr_xy_gp( const tjob * job, const OSTATE * state, 
                       const double * chan[2], FILE * gp )
{
    double xdiv, xexp, ydiv, yexp;
    char xpre[2], ypre[2];

    /* prepare scales */
    neat_scale( state->Yv[0], &ydiv, &yexp, ypre );
    neat_scale( state->Yv[1], &xdiv, &xexp, xpre );
    /* write gnuplot file */
    gnuplot_header( job, state, TXY, gp );
    /* x axis */
    fprintf( gp, "set xrange [%.1f:%.1f]\n",  -5*xdiv, 5*xdiv );
    fputs( "set mxtics 2\n", gp );
    fprintf( gp, "set xtics %.1f\n",  2*xdiv );
    fprintf( gp, "set xtics add ('%sV' %.1f 2)\n", xpre, 3*xdiv );
    /* y axis */
    fprintf( gp, "set yrange [%.0f:%.0f]\n", -4*ydiv, 4*ydiv );
    fprintf( gp, "set ytics %.0f\n",  ydiv );
    fprintf( gp, "set ytics add ('%sV' %f 3)\n", ypre, 3.5*ydiv );
    fputs( "\n", gp );  
    /* plot */
    fprintf( gp, "plot '-' using ($1*%e):($2*%e) notitle with points\n", 
             xexp, yexp );  
    /* print data */
    print_xy( gp, chan[1], chan[0], state->Yv[1], state->Yv[0] );
    fputs( "e\n", gp );
}

/**
 * Deliver chan and/or spec data to file or gnuplot 
 * As gnuplot shall be reused, it is given as gp.
 * Interactive: Channel, Spectrum, Correlation
 *              Ps, Eps, Gnuplot, pNg, Data, prinT
 *              after read (RETURN), File
 */
void job_deliver( const tjob * job, const OSTATE * state, 
                  const double * chan[2], const double * spec[2], 
		  const double * corr, FILE * gp )
{
    FILE * o;

    switch ( job->format )
    {
	case 'x': /* X11 */
	case 'e': /* eps */
	case 'p': /* ps */
	case 'n': /* png */
	    if ( job->ch ) 
	    {
		if ( 'y' == resolve_source( job, state ) ) 
		{
		    fputs( "Writing xy-plot.\n", stderr );
		    job_dlvr_xy_gp( job, state, chan, gp );
		}
		else
		{
		    fputs( "Writing channel data.\n", stderr );
		    job_dlvr_chan_gp( job, state, chan, gp );
		}
	    }
	    if ( job->sp ) 
	    {
		fputs( "Writing spectum.\n", stderr );
		job_dlvr_spec_gp( job, state, spec, gp );
	    }
	    if ( job->co ) 
	    {
		fputs( "Writing correlation.\n", stderr );
		job_dlvr_corr_gp( job, state, corr, gp );
	    }
	    fflush( gp );
	    break;
        case 'g': /* plt file */
	    if ( job->ch )
	    {
		if ( 'y' == resolve_source( job, state ) ) 
		{
		    fputs( "Writing XY-plot.\n", stderr );
		    o = open_outfile( job, TXY );
		    job_dlvr_xy_gp( job, state, chan, o );
		    close_outfile( job, o );
		}
		else
		{
		    fputs( "Writing channel data.\n", stderr );
		    o = open_outfile( job, TCHAN );
		    job_dlvr_chan_gp( job, state, chan, o );
		    close_outfile( job, o );
		}
	    }
	    if ( job->sp )
	    {
		fputs( "Writing spectum.\n", stderr );
		o = open_outfile( job, TSPEC );
		job_dlvr_spec_gp( job, state, spec, o );
		close_outfile( job, o );
	    }
	    if ( job->co )
	    {
		fputs( "Writing correlation.\n", stderr );
		o = open_outfile( job, TCORR );
		job_dlvr_corr_gp( job, state, corr, o );
		close_outfile( job, o );
	    }
	    break;
        case 'd': /* data file */
	    if ( job->ch )
	    {
		if ( 'y' == resolve_source( job, state ) )
		{
		    fputs( "Writing xy-plot.\n", stderr );
		    o = open_outfile( job , TXY ); 
		    job_dlvr_xy_data( state, chan, o );
		    close_outfile( job, o );
		}
		else
		{
		    fputs( "Writing channel data.\n", stderr );
		    o = open_outfile( job , TCHAN ); 
		    job_dlvr_chan_data( job, state, chan, o );
		    close_outfile( job, o );
		}
	    }
	    if ( job->sp )
	    {
		fputs( "Writing spectum.\n", stderr );
		o = open_outfile( job, TSPEC );
		job_dlvr_spec_data( job, state, spec, o );
		close_outfile( job, o );
	    }
	    if ( job->co )
	    {
		fputs( "Writing correlation.\n", stderr );
		o = open_outfile( job, TCORR );
		job_dlvr_corr_data( job, state, corr, o );
		close_outfile( job, o );
	    }
	    break;
	default:
	    fputs( "Unknown out format in job_deliver. It's a bug!\n", stderr );
	    exit( 1 );
    }
}


/**********************************************************************
			       data acquisition
 **********************************************************************/

/**
 * calculate correlation
 * Interacitve: coRrealtion
 */
void job_acq_corr( const tjob * job, const OSTATE * state, 
                   double * chan[2], double * corr )
{
  switch ( resolve_source( job, state ) )
  {
    case '1':
    case 'a':
      chan_corr( corr, chan[0], chan[0] );
      break;
    case '2':
      chan_corr( corr, chan[1], chan[1] );
      break;
    case 'b':
      chan_corr( corr, chan[0], chan[1] );
      break;
  }
}


/**
 * calculate spectrum
 * Interacitve: Spectrum
 */
void job_acq_spec( const tjob * job, const OSTATE * state, 
                   double * chan[2], double * spec[2] )
{
  if ( index( "1ba", resolve_source( job, state ) ) )
    chan_spec( chan[0], spec[0] );
  if ( index( "2b", resolve_source( job, state ) ) )
    chan_spec( chan[1], spec[1] );
}


/** 
 * get state from file
 * acquire data from file
 * return: 1 on success, 0 on error
 */
int job_acq_file( const tjob * job, OSTATE * state, double * chan[2] )
{
  FILE * in;
  double Ts2;

  /* init state */
  state->Yon[0] = 1;   state->Yon[1] = 0;
  state->Ys[0]  = 1;   state->Ys[1]  = 1;
  state->Yc[0]  = 'A'; state->Yc[1]  = 'A';
  state->add = 0;
  state->XY = 0;
  state->ok = 1;
  /* open file */
  in = fopen( job->in, "r" );
  if ( ! in )
  {
    fprintf( stderr, "Can't open file %s.\n", job->in );
    /* hence fclose( in ); is not possible */
    return 0;
  }
  /* read channels */
  if ( ! read_chan( in, chan[0], &state->Ts, &state->Yv[0]  ) )
  {
    fprintf( stderr, "Something wrong with file %s.\n", job->in );
    fclose( in );
    return 0;
  }
  if ( read_chan( in, chan[1], &Ts2, &state->Yv[1]  ) )
  {
    state->Yon[1] = 1;
    if ( Ts2 != state->Ts )
      fprintf( stderr, 
  "WARNIG! Timescales differ! (%6.0e s/DIV and %6.0e s/DIV) Using the first.\n",
      state->Ts, Ts2 );
  }
  /* clean up */
  fclose( in );
  return 1;
}


/** 
 * get state from scope
 * acquire data from scope
 * return: 1 on success, 0 on error
 */
int job_acq_scope( const tjob * job, OSTATE * state, double * chan[2] )
{
    OSC * osc;
    int i;

    /* open scope */
    osc = osc_open( job->serial );
    if ( osc->error[0] != 0 )
    {
	fprintf( stderr, "%s\n", osc->error );
	return 0;
    }
    osc_get_state( osc, state );
    if ( ! state->ok )
    {
	fputs( "Scope has to be in storage mode, Timebase A.\n", stderr );
	osc_close( osc );
	return 0;
    }
    /* set to hold, if not alredy set 
     * The HLDWFM? query does not work correctly: 
     * - it always returns '0'
     * - after the query, the reading of channels is not reliable.
	int c; 
	osc_write_str( osc, "HLDWFM?" );
	osc_read_str( osc, "HLDWFM:", &c, 1 );
        printf( "c=%c\n", c );
    */
    osc_write_str( osc, "HLDWFM=1" );
    osc_get_returncode( osc );
    /* read channels, apply probe factor */
    if ( index( "1bay", resolve_source( job, state ) ) )
    {
	osc_get_chan( osc, state, 1, chan[0] );
	for ( i = 0; i < 2048; i++)
	    chan[0][i] *= job->probes[0];
	state->Yv[0] *= job->probes[0];
    }
    if ( index( "2by", resolve_source( job, state ) ) )
    {
	osc_get_chan( osc, state, 2, chan[1] );
	for ( i = 0; i < 2048; i++)
	    chan[1][i] *= job->probes[1];
	state->Yv[1] *= job->probes[1];
    }
    /* close scope */
    osc_write_str( osc, "HLDWFM=0" );
    osc_get_returncode( osc );
    osc_close( osc );
    return 1;
}


/** 
 * Read data from file or scope, calc spec and/or corr.
 * Interactive: RETURN (data acquisition), File
 * return: 1 on success, 0 on error
 */
int job_acquire( const tjob * job, OSTATE * state, 
                  double * chan[2], double * spec[2], double * corr )
{
    if ( job->source == 'f' )
    {
        fputs( "Reading from file.\n", stderr );
        if ( ! job_acq_file( job, state, chan ) )
            return 0;
    }
    else
    {
        fputs( "Reading from scope.\n", stderr );
        if ( !job_acq_scope( job, state, chan ) )
            return 0;
    }
    if ( job->sp )
    {
        if ( 'y' == resolve_source( job, state ) )
        {
	    fputs( "No spectrum in XY mode.\n", stderr );
	    return 0;
        }
        fputs( "Calculating spectrum.\n", stderr );
        job_acq_spec( job, state, chan, spec );
    }
    if ( job->co )
    {
        if ( 'y' == resolve_source( job, state ) )
        {
	    fputs( "No correlation in XY mode.\n", stderr );
	    return 0;
        }
        fputs( "Calculating correlation.\n", stderr );
        job_acq_corr( job, state, chan, corr );
    }
    return 1;
}


/**********************************************************************
		      job management -- user interface
 **********************************************************************/

/**
 * reads one char unbuffered from stdin
 *
 * returns -1 on error
 */
char readchar()
{
  char ch;
  int n;
  static struct termios Otty, Ntty;

  fflush( stdout );
  tcgetattr( 0, &Otty );
  Ntty = Otty;
  /* unbuffered settings */
  Ntty.c_iflag = 0;
  Ntty.c_oflag = 0;
  Ntty.c_lflag &= ~ICANON; /* switch off line buffer */
  Ntty.c_lflag |= ECHO;  /* enable echo */
  Ntty.c_cc[VMIN]  = 1; /* minimum chars to wait for */
  Ntty.c_cc[VTIME] = 255; /* minimum wait time (ds). Does value matter? */
  if ( 0 != tcsetattr( 0, TCSAFLUSH, &Ntty) )
    return -1;
  n = read( 0, &ch, 1 ); 
  if ( 0 != tcsetattr( 0, TCSAFLUSH, &Otty) )
    return -1;
  if ( n != 1 )
    return -1;
  return (int)ch;
}


/**
 * read a line from f to buffer b with length l
 * If string in b ends on NL, replace it by \0.
 */
char * readline( FILE * f, char * b, size_t l )
{
  char * r;
  size_t n;

  r = fgets( b, l, f );
  if ( r ) 
  {
    n = strlen( b ) - 1;
    if ( b[n] == '\n') 
      b[n] = '\0';
  }
  return r;
}


/**
 * check if fname is suitable for autonumbering
 *
 * To make autonumbering possible, a substring like /%[:digits:]*d/ has to be
 * present. No further % signs may be present.
 *
 * returns 0 -> no autonumbering; 1 -> autonumbering wanted
 */
int has_autonum( const char * fname )
{
    char * p;

    /* find first % sign */
    p = strchr( fname, '%' );
    if ( ! p ) 
	return 0;
    /* count number of subsequent numbers */
    p++;
    p += strspn( p, "0123456789" );
    /* check for 'd' */
    if ( *p != 'd' )
	return 0;
    /* check for further % signs */
    p++;
    p = strchr( p, '%' );
    if ( p != NULL )
	return 0;
    return 1;
}


/**
 * gnuplot greeting screen
 *
 * hopefully assures mouse function
 */
void int_welcome( FILE * gp )
{
  fputs( "set mouse\n", gp );
  fputs( "set terminal x11 0 title 'Monitor HM'\n", gp );
  fputs( "set format ''\n", gp );
  fputs( "set zeroaxis\n", gp );
  fputs( "set label 'HMacquire' at 0, 0.3 center\n", gp );
  fputs( "\n", gp );
  fputs( "plot [-3.1415:3.1415][-1.1:1.1] sin(x) notitle\n", gp );
  fflush( gp );
}


/**
 * Print selection box with braces.
 * 
 * The string s is printed, 
 * either in [] for ( a == b )
 * or in spaces for ( a != b )
 */
void int_print_selection( FILE * f, const char * s, int a, int b )
{
  if ( a == b ) 
    fprintf( f, "\e[1m[%s]\e[m", s ); /* ANSII ESC */
  else
    fprintf( f, " %s ", s );
}


/**
 * Express source as written string
 *
 * s  string to write to
 * n  length of string buffer
 */
void int_source_to_str( char * s, size_t n, const tjob * job, 
                     const OSTATE * state )
{
  char source;

  source = job->source;
  if ( job->source == 'o' )
    source = resolve_source( job, state );
  switch ( source )
  {
    case '1':
    case '2':
      snprintf( s, n, "CH%c", source );
      break;
    case 'b':
      strncpy( s, "CH1 & CH2", n );
      break;
    case 'a':
      snprintf( s, n, "%cCH1 %c CH2", 
               (state->Ys[0] == 1)?'+':'-', (state->Ys[1] == 1)?'+':'-' );
      break;
    case 'y':
      snprintf( s, n, "XY" );
      break;
    case 'f':
      strncpy( s, job->in, n );
      break;
  }
  s[n-1] = 0; /* assure 0 terminated string */
}


/**
 * print command prompt
 */
void int_print_prompt( FILE * f, const tjob * job, const char * seen_now, 
                       const char * message )
{
  fputs( "\e[H\e[J", f ); /* ANSII ESC */
  fprintf( f, "message  :  %s\n", message );
  fprintf( f, "seen now :  %s\n\n", seen_now );
  /* source line*/
  fputs( "source   : ", f );
  int_print_selection( f, "1", job->source, '1' );
  int_print_selection( f, "2", job->source, '2' );
  int_print_selection( f, "Both", job->source, 'b' );
  int_print_selection( f, "Add", job->source, 'a' ),
  int_print_selection( f, "xY", job->source, 'y' ),
  int_print_selection( f, "scOpe", job->source, 'o' );
  fputs( "\n", f );
  /* probes line*/
  fprintf( f, "probes   :  Ch1 x%4g                cH2 x%4g\n", 
	      job->probes[0], job->probes[1] );
  /* display line*/
  fputs( "display  : ", f );
  int_print_selection( f, "Time-domain", job->ch, 1 ),
  int_print_selection( f, "Spectrum", job->sp, 1 );
  int_print_selection( f, "coRrelation", job->co, 1 );
  fputs( "\n", f );
  /* remaining lines */
  fputs( "save as  :  Ps   Eps    pNg    Gnuplot    Data\n", f );
  fputs( "read from:  <-' source                    File\n", f );
  fputs( "            prInt                         Quit\n", f );
  fputs( "\nHMacquire > ", f );
  fflush( f );
}


/**
 * Ask for string
 *
 * give prompt "q [buf]> "
 * On empty input buffer is unchanged.
 * assumed buffer length is INTBUFLEN.
 *
 * q    question
 * buf  buffer to read to
 * returns 0 -> buf not changed; 1 -> buf changed
 */
int int_ask_str( const char * q, char * buf )
{
  char line[INTBUFLEN] = "";

  fprintf( stdout, "%s [%s]> ", q, buf );
  fflush( stdout );
  readline( stdin, line, INTBUFLEN );
  if ( line[0] )
    strcpy( buf, line );
  return line[0] != 0;
}


/**
 * Ask for float number
 *
 * give prompt "q [buf]> "
 * On empty input float is unchanged.
 *
 * q    question
 * f    double var to read to
 * returns 0 -> number not changed; 1 -> number changed
 */
int int_ask_float( const char * q, double * f )
{
    char line[INTBUFLEN] = "";
    double oldf = *f;

    fprintf( stdout, "%s [%g]> ", q, *f );
    fflush( stdout );
    readline( stdin, line, INTBUFLEN );
    if ( line[0] )
	*f = atof( line );
    return oldf != *f;
}


/**
 * main loop of interactive mode
 */
void interactive( tjob * job, OSTATE * state, 
                  double * chan[2], double * spec[2], double * corr )
{
    FILE * gp;
    char c;
    char seen_now[INTBUFLEN] = "";
    char message[INTBUFLEN] = "OK";
    char infile[INTBUFLEN] = "";
    char outfile[INTBUFLEN] = "";
    char printcmd[INTBUFLEN+1] = LPR;
    char backup, ch_bak, sp_bak, co_bak;
    int str_changed;
    int autonum;
    char autonum_buf[INTBUFLEN];

    /* init */
    gp = open_gnuplot();
    int_welcome( gp );
    /* respect command line filenames */
    if ( job->in ) strncpy( infile, job->in, INTBUFLEN );
    if ( job->out ) strncpy( outfile, job->out, INTBUFLEN );
    infile[INTBUFLEN-1] = outfile[INTBUFLEN-1] = 0;
    job->in = job->out = NULL;
    /* main loop */
    do 
    {
	int_print_prompt( stdout, job, seen_now, message );
	strcpy( message, "OK" );
	c = readchar();
	fputs( "\n", stdout );
	switch ( c )
	{
	    case '1': /* source 1     */
	    case '2': /* source 2     */ 
	    case 'b': /* source Both  */ 
	    case 'a': /* source Add   */ 
	    case 'o': /* source scOpe */ 
		job->source = c;
		break;
	    case 'y': /* source xY    */
		job->source = c;
		job->ch = 1;
		job->sp = 0;
		job->co = 0;
		fputs( "set terminal x11 1 close\n", gp );
		fputs( "set terminal x11 2 close\n", gp );
		break;
	    case 'c': 
		int_ask_float("Factor for CH1 probe?", &job->probes[0]);
		break;
	    case 'h': 
		int_ask_float("Factor for CH2 probe?", &job->probes[1]);
		break;
	    case 't':	/* display Time-domain */
		job->ch = 1 - job->ch;
		if ( job->ch )
		{
		    if ( seen_now[0] == 0 )
			break; /* if never read data, dont calc spectrum now */
		    sp_bak = job->sp; job->sp = 0;
		    co_bak = job->co; job->co = 0;
		    job->format = 'x';
		    job_deliver( job, state, (const double **)chan, 
		                             (const double **)spec, corr, gp );
		    job->sp = sp_bak;
		    job->co = co_bak;
		}
		else
		{
		    fputs( "set terminal x11 0 close\n", gp );
		    fflush( gp );
		}
		break;
	    case 's': /* display Spectrum */
		if ( 'y' == resolve_source( job, state ) )
		{
		    strcpy( message, "Not in XY mode." );
		    break;
		}
		job->sp = 1 - job->sp;
		if ( job->sp )
		{
		    if ( seen_now[0] == 0 )
			break; /* if never read data, dont calc spectrum now */
		    job_acq_spec( job, state, chan, spec );
		    ch_bak = job->ch; job->ch = 0;
		    co_bak = job->co; job->co = 0;
		    job->format = 'x';
		    job_deliver( job, state, (const double **)chan, 
					     (const double **)spec, corr, gp );
		    job->ch = ch_bak;
		    job->co = co_bak;
		}
		else
		{
		    fputs( "set terminal x11 1 close\n", gp );
		    fflush( gp );
		}
		break;      
	    case 'r': /* display correlation */
		if ( 'y' == resolve_source( job, state ) )
		{
		    strcpy( message, "Not in XY mode." );
		    break;
		}
		job->co = 1 - job->co;
		if ( job->co )
		{
		    if ( seen_now[0] == 0 )
			/* if never read data, dont calc correlation now */
			break; 
		    job_acq_corr( job, state, chan, corr );
		    ch_bak = job->ch; job->ch = 0;
		    sp_bak = job->sp; job->sp = 0;
		    job->format = 'x';
		    job_deliver( job, state, (const double **)chan, 
					     (const double **)spec, corr, gp );
		    job->ch = ch_bak;
		    job->sp = sp_bak;
		}
		else
		{
		    fputs( "set terminal x11 2 close\n", gp );
		    fflush( gp );
		}
		break;
	    case  '\r' : /* read from source */
	    case  '\n' :
		/* acquire data */
		if ( ! job_acquire( job, state, chan, spec, corr ) ) 
		{
		    snprintf( message, INTBUFLEN, "Can't read scope at '%s'.", 
		    job->serial );
		    message[INTBUFLEN-1] = 0;
		    break;
		}
		int_source_to_str( seen_now, INTBUFLEN, job, state );
		/* update display */
		job->format = 'x';
		job_deliver( job, state, (const double **)chan, 
					 (const double **)spec, corr, gp );
		/* autonumbering: increase number */
		autonum++;
		break;
	    case 'f': /* read from File */
		/* ask filename */
		int_ask_str( "Enter filename", infile );
		if ( ! infile[0] )
		{
		    strcpy( message, "No filename given." );
		    break;
		}
		/* acquire data */
		backup = job->source;
		job->source = 'f';
		job->in = infile;
		if ( job_acquire( job, state, chan, spec, corr ) )
		{
		    int_source_to_str( seen_now, INTBUFLEN, job, state );
		    /* update display */
		    job->format = 'x';
		    job_deliver( job, state, (const double **)chan, 
					     (const double **)spec, corr, gp );
		    /* autonumbering: increase number */
		    autonum++;
		}
		else
		{
		    snprintf( message, INTBUFLEN, "Can't read '%s'.", infile );
		    message[INTBUFLEN-1] = 0;
		}
		job->source = backup;
		job->in = NULL;
		break;
	    case 'e': /* save as Eps     */
	    case 'p': /* save as Ps      */
	    case 'n': /* save as pNg     */
	    case 'g': /* save as Gnuplot */
	    case 'd': /* save as Data    */
		/* ask filename */
		str_changed = int_ask_str( "Enter basename", outfile );
		if ( ! outfile[0] )
		{
		    strcpy( message, "No filename given." );
		    break;
		}
		job->out = outfile;
		/* autonumbering */
		if ( str_changed ) 
		    autonum = 0;
		if ( has_autonum( outfile ) )
		{
		    snprintf( autonum_buf, INTBUFLEN, outfile, autonum );
		    job->out = autonum_buf;
		}
		/* deliver data */
		snprintf( message, INTBUFLEN, "Writing '%s'", job->out );
		job->format = c;
		job_deliver( job, state, (const double **)chan, 
					 (const double **)spec, corr, gp );
		job->out = NULL;
		break;
	    case 'i': /* prInt */
		/* ask print command */
		int_ask_str( "Enter print cmd", &printcmd[1] );
		/* deliver data */
		job->out = printcmd;
		job->format = 'p';
		job_deliver( job, state, (const double **)chan, 
					 (const double **)spec, corr, gp );
		job->out = NULL;
		sleep( 2 );
	    break;
	    case 'q': /* Quit */
		break;
	    default:
		sprintf( message, "Unknown command '%c'.", c );
	}
    }
    while ( c != 'q' );
    /* clean up */
    close_gnuplot( gp );
}


/**
 * manage noninteractive job
 */
void do_job( const tjob * job, OSTATE * state, 
             double * chan[2], double * spec[2], double * corr )
{
    FILE * gp;

    if ( ! job_acquire( job, state, chan, spec, corr ) )
	exit( 1 );
    if ( index( "xepn", job->format ) )
	gp = open_gnuplot();
    else
	gp = stdout;
    job_deliver( job, state, (const double **)chan, (const double **)spec, 
                 corr, gp );
    if ( 'x' == job->format )
    {
	fputs( "set terminal x11 0 persist\n", gp );
	fputs( "set terminal x11 1 persist\n", gp );
	fputs( "set terminal x11 2 persist\n", gp );
    }
    if ( index( "xepn", job->format ) )
	close_gnuplot( gp );
    /* Frage 5.7: Warum kann ich keinen "char **" an ein Funktion übergeben, die
     * einen "const char **" erwartet?
     *
     * Antwort: Zwar kann man stets zeiger-auf-T (für alle Typen T) verwenden wo
     * zeiger-auf-const-T erwartet wird. Diese Ausnahmeregel gilt aber nicht
     * rekursiv!
     *
     * Für Zuweisungen zwischen Zeigern, deren Qualifier sich auf einer anderen
     * als der ersten Stufe unterscheiden, muss eine explizite Typkonvertierung
     * (z.B. in diesem Fall "(const char **)") verwendet werden.
     *
     * References: ANSI Sec. 3.1.2.6, Sec. 3.3.16.1, Sec. 3.5.3; ISO Sec.
     * 6.1.2.6, Sec. 6.3.16.1, Sec. 6.5.3; H&S Sec. 7.9.1 pp. 221-222. 
     * 
     * Quelle: FAQ aus de.comp.lang.c
     * http://www2.informatik.uni-wuerzburg.de/dclc-faq/kap5.html
     */
}


/**********************************************************************
		       user interface for main program
 **********************************************************************/


/** Print help and exit with n */
void help( int n )
{
    fputs( 
	"hmacquire [jobspec] [options]\n"
	"\n"
	"jobspec may contain\n"
	"  one source:\n"
	"    1 - channel 1\n"
	"    2 - channel 2\n"
	"    b - both channels\n"
	"    a - add both channels\n"
	"    y - xy plot\n"
	"    o - take active source from oscilloscope\n"
	"    f - read from datafile (see format below)\n"
	"  one format:\n"
	"    x - x11 window\n"
	"    e - eps file\n"
	"    p - ps file\n"
	"    n - png file\n"
	"    g - gnuplot file (commands and data)\n"
	"    d - data file (may be source again)\n"
	"  at least one item to display:\n"
	"    c - channel data\n"
	"    s - spectrum\n"
	"    r - correlation\n"
	"\n"
	"options\n"
	"  -o <basename>   basename for outputfiles, otherwise stdout\n"
	"  -f <inputfile>  inputfile for source f, otherwise stdin\n"
	"  -s <device>     device for serial port, otherwise " SERIALPORT "\n"
	"  -i              ineractive mode\n"
	"  -v              print program version and copyright statement\n"
	"  -v <device>     print firmware version for scope at port device\n"
	"  -b              Send break to serial port before communication.\n"
	"                  Multiple -b send multiple breaks.\n"
	"  -pN <f>         Apply voltage probe factor f to channel N.\n"
	"                  For 1:10 probe on channel 1 use '-p1 10'\n"
	"                  Default: 1 for both channels\n"
	"\n"
	"Interactive mode\n"
	"----------------\n"
	"Use capitalized letters to select functions.\n"
	"\n"
	"The 'save as' function allows for automatic numbering. To use it,\n"
	"place the string '%' numbers 'd' in the file name (e. g. '%05d').\n"
	"The pattern will be substituted by a consecutive number according\n"
	"to the rules of printf. Only one '%' sign is allowed. The number\n"
	"becomes increased on a new measurement and reset on a change of the\n"
	"filename.\n"
	"\n"
	"\n", stderr );
    exit( n );
}


/** Print error message and exit with code 1 */
void error( const char * e )
{
  fputs( e, stderr );
  fputs( "\nHelp with -h.\n", stderr );
  exit( 1 );
}


/** Print scope version to stderr */
void version( const char * port )
{
  OSC * osc;

  osc = osc_open( port );
  if ( osc->error[0] ) 
  {
    fprintf( stderr, "Error: %s\n", osc->error );
    exit( 1 );
  }
  else
    fprintf( stderr, "Oscilloscope Version: %s\n", osc->version );
  osc_close( osc );
  exit( 0 );
}


/** Print program version and copyright */
void copyleft()
{
    fputs( 
	"HM acquire - software for reading out HAMEG digital storage\n"
	"oscilloscopes. Version " VERSION "\n\n"
	"Copyright (C) 2007, 2011 Lars Stollenwerk\n"
	"\n"
	"This program is free software; you can redistribute it and/or\n"
	"modify it under the terms of the GNU General Public License as\n"
	"published by the Free Software Foundation; either version 2 of\n"
	"the License, or (at your option) any later version.\n"
	"\n"
	"This program is distributed in the hope that it will be\n"
	"useful, but WITHOUT ANY WARRANTY; without even the implied\n"
	"warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR\n"
	"PURPOSE. See the GNU General Public License for more\n"
	"details.\n"
	"\n"
	"You should have received a copy of the GNU General\n"
	"Public License along with this program; if not, write\n"
	"to the Free Software Foundation, Inc., 51 Franklin St,\n"
	"Fifth Floor, Boston, MA 02110, USA\n"
	"\n"
	"Contact the author via\n"
	"LSto@gmx.de\n"
	"http://formica.nusseis.de/HMacquire/\n", stderr );
    exit( 0 );
}


/** Read command line options, write to job */
void read_args( int argc, char ** argv, tjob * job )
{
  int i,j,N;
  int serial_break = 0;
  int serial_version = 0;

  /* init */
  job->source = 'o';
  job->ch = 0;
  job->sp = 0;
  job->co = 0;
  job->format = 'x';
  job->out = NULL;
  job->in = NULL;
  job->serial = SERIALPORT;
  job->interactive = 0;
  job->probes[0] = job->probes[1] = 1;
  /* read args */
  for ( i = 1; i < argc; i++ )
    if ( argv[i][0] == '-' )
      switch ( argv[i][1] )
      {
	case 'h':
	  help( 0 );
	case 'v':
	  if ( ++i == argc )
	    copyleft();
	  else 
	  {
	    job->serial = argv[i];
	    serial_version = 1;
	  }
	  break;
	case 'i':
	  job->interactive = 1;
	  break;
        case 'o':
	  if ( ++i == argc ) error( "Name a basename for the output files." );
	  job->out = argv[i];
	  break;
        case 'f':
	  if ( ++i == argc ) error( "Name an inputfile." );
	  job->in = argv[i];
	  break;
        case 's':
	  if ( ++i == argc ) error( "Name a device for serial port." );
	  job->serial = argv[i];
	  break;
	case 'b':
	  serial_break++;
	  break;
	case 'p':
	  N = atoi(&argv[i][2]);
	  if ( N != 1 && N != 2 ) error("Channel must be 1 or 2.");
	  N--;
	  if ( ++i == argc ) error( "Give a probe factor." );
	  job->probes[N] = atof(argv[i]);
	  if (job->probes[N] == 0) error("Give finite probe factor as float.");
	  break;
        default:
	  error( "Unknown option." );
      }
    else
      for ( j = 0; j < strlen( argv[i] ); j++)
        switch ( argv[i][j] )
	{
	  case '1': case '2': case 'b': case 'a': case 'y': case 'o': case 'f':
	    job->source = argv[i][j];
	    break;
	  case 'x': case 'e': case 'p': case 'n': case 'g': case 'd': 
	    job->format = argv[i][j];
	    break;
	  case 'c':
	    job->ch = 1;
	    break;
	  case 's':
	    job->sp = 1;
	    break;
	  case 'r':
	    job->co = 1;
	    break;
	  default:
	    error( "Unknown command letter." );
	}
  /* send serial break */
  for ( i = 0; i < serial_break; i++ )
    osc_break( job->serial );
  /* print scope version */
  if ( serial_version )
    version( job->serial );
  /* check consistency */
  if ( job->interactive )
  {
    if ( !job->ch && !job->sp && !job->co )
      job->ch = 1;
    if ( job->source == 'f' )
      error( "Choose reading from file from menue." );
    if ( job->format != 'x' )
      error( "Choose output format from menue." );
  }
  else /* non-interactive */
  {
    if ( !job->ch && !job->sp && !job->co )
      error( "You want a Channel? Or a spectrum? Or both?" );
    if ( job->in && job->source != 'f'  )
      error( "Shall I read from file?" );
    if ( job->out && !index( "egdp", job->format ) )
      error( "There'd be nothig to write to file." );
  }
}


/** display job inforation for user */
void print_job( const tjob * job, FILE * f )
{
  fprintf( f, "Source   : %c\n", job->source );
  fprintf( f, "Format   : %c\n", job->format );
  if ( job->ch )
    fputs( "Output   : channel data\n", f );
  if ( job->sp )
    fputs( "Output   : spectrum\n", f );
  if ( job->co )
    fputs( "Output   : correlation\n", f );
  fprintf( f, "Outfile  : %s\n", job->out?job->out:"STDOUT" );
  if ( job->source == 'f' )
    fprintf( f, "Inputfile: %s\n", job->in?job->in:"STDIN" );
  else
    fprintf( f, "Scope at : %s\n", job->serial );
  fprintf( f, "Voltage probe CH1: %g\n", job->probes[0] );
  fprintf( f, "Voltage probe CH2: %g\n", job->probes[1] );
}


/**********************************************************************
				 main program
 **********************************************************************/


/** main program */
int main ( int argc, char ** argv )
{
  tjob job;
  OSTATE state;
  double * chan[2];
  double * spec[2];
  double * corr;

  /* read commandline */
  read_args( argc, argv, &job );
  print_job( &job, stderr );
  /* allocate data memory */
  chan[0] = malloc( 2 * 2048 * sizeof(double) );
  chan[1] = &chan[0][2048];
  spec[0] = malloc( 2 * 1025 * sizeof(double) );
  spec[1] = &spec[0][1025];
  corr = malloc( 4096 * sizeof(double) );
  /* do the job */
  if ( job.interactive )
    interactive( &job, &state, chan, spec, corr );
  else
    do_job( &job, &state, chan, spec, corr );
  /* clean up */
  free( chan[0] );
  free( spec[0] );
  free( corr );
  return 0;
}


/* vim: set fdm=expr foldexpr=getline(v\:lnum)=~'^{'?'>1'\:getline(v\:lnum)=~'^}'?'<1'\:'-1' : */
