/**
 * HM acquire - software for reading out HAMEG digital storage
 * oscilloscopes.
 *         
 * Copyright (C) 2007 -- 2018 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@opppf.de
 * http://opppf.de/HMacquire/
 */


#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "osc_api.h"
#include "hm_constants.h"


/* ----------------------------------------------------------------------
			      internal functions
 ---------------------------------------------------------------------- */


/**
 * Read n bytes from scope.
 *
 * o       OSC struct
 * b       buffer to write to
 * m       number characters to read
 */
void osc_read_bytes( OSC * o, char * b, int m )
{
  int n = 0;
  int ret;

  do {
    ret = read( o->fd, &b[n], m-n ); 
    if ( ret < 0 )
    {
      strerror_r( errno, o->error, 256 );
      fprintf( stderr, "%s\n", o->error );
    }
    n += ret;
  } while ( n < m );
}


/* ----------------------------------------------------------------------
				api functions
 ---------------------------------------------------------------------- */


/**
 * Open connection to scope.
 *
 * c	   filename of serial port
 * return: OSC descriptor
 *         possible error in field error.
 */
OSC * osc_open( const char * c )
{
  OSC * o;
  int e;
  struct termios options;
  
  o = malloc( sizeof(OSC) );
  /* init error detection */
  o->error[0] = 0; 
  e = 0;
  /* open port */
  o->fd = open( c, O_RDWR | O_NOCTTY );
  if ( o->fd == -1 ) 
  { 
    strerror_r( errno, o->error, 256 );
    return o; 
  }
  /* get port options */
  if ( tcgetattr( o->fd, &options ) == -1 )
  { 
    strerror_r( errno, o->error, 256 );
    close( o->fd );
    return o; 
  }
  /* baud rate */
  e += cfsetispeed( &options, B19200 );
  e += cfsetospeed( &options, B19200 );
  /* select 8N2 */
  options.c_cflag &= ~CSIZE; 
  options.c_cflag |= CS8;   
  options.c_cflag |= CSTOPB;   
  options.c_cflag &= ~PARENB;
  /* RTS/CTS handshake */
  options.c_cflag |= CRTSCTS;   
  /* raw in- and output */
  options.c_iflag |= IGNBRK | IGNPAR;
  options.c_iflag &= ~(BRKINT|PARMRK|INPCK|ISTRIP|INLCR|IGNCR|ICRNL|IXON|IXOFF);
  options.c_oflag &= ~( OPOST | ONLCR | OCRNL | ONOCR | ONLRET | OFILL );
  options.c_cflag |= ( CLOCAL | CREAD | HUPCL );
  options.c_lflag &= ~( ICANON | ECHO | ECHOE | ISIG );
  /* set port options */
  e += tcsetattr( o->fd, TCSANOW, &options );
  /* init scope: SPACE CR */
  if ( write( o->fd, " \r", 2 ) != 2 ) { e += -1; }
  /* Check setting of serial port */
  if ( e != 0 ) 
  { 
    strcpy( o->error, "Can't set up serial port." ); 
    close( o->fd );
    return o; 
  }
  /* check scope answer */
  if ( osc_get_returncode( o ) != '0' ) 
  { 
    strcpy( o->error, "Scope doesn't answer." ); 
    close( o->fd );
    return o; 
  }
  /* read firmware version */
  osc_write_str( o, "VERS?" );
  if ( osc_read_str( o, "VERS:", o->version, 15 ) )
  {
    o->version[13] = 0;
    strncpy( o->ver, &o->version[2], 4 );
    o->ver[4] = 0;
  }
  else
  {
    strcpy( o->error, "Can't read firmware version." ); 
    close( o->fd );
    return o; 
  }
  osc_write_str( o, "LK=1" );
  osc_get_returncode( o );
  return o;
}


/**
 * Close Connection to scope.
 *
 * o	OSC struct
 */
void osc_close( OSC * o )
{
  osc_write_str( o, "RM0" );
  osc_get_returncode( o );
  close( o->fd );
  free( o );
}


/**
 * Write 0 terminated string to scope.
 *
 * o	OSC struct
 * c	string to write
 * return: 1 = success; 0 = error
 */
int osc_write_str( OSC * o, const char * c )
{
  int e1, e2;

  /* clear errror message */
  o->error[0] = 0;
  e1 = write( o->fd, c, strlen( c ) );
  e2 = write( o->fd, "\r\n", 2 );
  if ( e1 == -1 || e2 == -1 )
  {
    strcpy( o->error, "osc_write_str" ); 
    return 0;
  }
  return 1;
}


/**
 * Write string with n bytes to scope.
 *
 * o	OSC struct
 * c	string to write
 * n	length of string
 * return: 1 = success; 0 = error
 */
int osc_write_strn( OSC * o, const char * c, int n )
{
  int e1, e2;

  /* clear errror message */
  o->error[0] = 0;
  e1 = write( o->fd, c, n );
  e2 = write( o->fd, "\r\n", 2 );
  if ( e1 == -1 || e2 == -1 )
  {
    strcpy( o->error, "osc_write_strn" ); 
    return 0;
  }
  return 1;
}


/**
 * Read 0 terminated string from scope and store m further bytes to b.
 *
 * o	OSC struct
 * c	string to receive
 * b	destination for answer
 * m	bytes of answer
 * return: 1 = success; 0 = error
 */
int osc_read_str( OSC * o, const char * c, void * b, int m )
{
  int n;
  char cb[256];

  /* clear errror message */
  o->error[0] = 0;
  /* read string part of answer */
  n = strlen( c );
  osc_read_bytes( o, cb, n );
  cb[n] = 0;
  if ( strcmp( cb, c ) != 0 )
  {
    strcpy( o->error, "osc_read_str" ); 
    return 0;
  }
  /* read data part of answer */
  osc_read_bytes( o, b, m );
  return 1;
}


/**
 * Read  string with n bytes from scope and store m further bytes to b.
 *
 * o	OSC struct
 * c	string to receive
 * n	length of string
 * b	destination for answer
 * m	bytes of answer
 * return: 1 = success; 0 = error
 */
int osc_read_strn( OSC * o, const char * c, int n, void * b, int m )
{
  char cb[256];

  /* clear errror message */
  o->error[0] = 0;
  /* read string part of answer */
  osc_read_bytes( o, cb, n );
  if ( memcmp( cb, c, n ) != 0 )
  {
    strcpy( o->error, "osc_read_strn" ); 
    return 0;
  }
  /* read data part of answer */
  osc_read_bytes( o, b, m );
  return 1;
}


/**
 * Read returncode R CR NL
 *
 * o	OSC struct
 * return: The R, 0x00 on error
 */
char osc_get_returncode( OSC * o )
{
  char b[3];

  osc_read_bytes( o, b, 3 );
  /* error treatment */
  o->error[0] = 0;
  /* Caution! No processing of CR and NL by serial kernal driver allowed. */
  if ( b[1] == '\r' && b[2] == '\n' ) return b[0];
  strcpy( o->error, "Can not read return code."  );
  return 0;
}


/**
 * Get scope state
 *
 * o	OSC struct
 * s	OSTATE struct
 * return: 1 = success; 0 = error
 */
int osc_get_state( OSC * o, OSTATE * s )
{
  unsigned char b;

  o->error[0] = 0; 
  /* channel 1 */
  osc_write_str( o, "CH1?" );
  osc_read_str( o, "CH1:", &b, 1 );
  if ( ( b & HM_Y_MASK ) > HM_Y_MAX )
  {
    strcpy( o->error, "Illegal Y1 amplification."  );
    return 0;
  }
  s->Yon[0] = b & HM_Y_ON;
  s->Yv[0] = HM_Y_V_DIV[ b & HM_Y_MASK ];
  s->Ys[0] = (b & HM_Y_INV)?-1:1;
  s->Yc[0] = (b & HM_Y_AC)?'A':'D';
  osc_write_str( o, "CH1VAR?" );
  osc_read_str( o, "CH1VAR:", &b, 1 );
  s->Yv[0] *= (double)(2*0xff) / (double)(b+0xff); /* scale up by max 2 */
  /* channel 2 */
  osc_write_str( o, "CH2?" );
  osc_read_str( o, "CH2:", &b, 1 );
  if ( ( b & HM_Y_MASK ) > HM_Y_MAX )
  {
    strcpy( o->error, "Illegal Y2 amplification."  );
    return 0;
  }
  s->Yon[1] = b & HM_Y_ON;
  s->Yv[1] = HM_Y_V_DIV[ b & HM_Y_MASK ];
  s->Ys[1] = (b & HM_Y_INV)?-1:1;
  s->Yc[1] = (b & HM_Y_AC)?'A':'D';
  osc_write_str( o, "CH2VAR?" );
  osc_read_str( o, "CH2VAR:", &b, 1 );
  s->Yv[1] *= (double)(2*0xff) / (double)(b+0xff); /* scale up by max 2 */
  /* vertical mode */
  osc_write_str( o, "VERMODE?" );
  osc_read_str( o, "VERMODE:", &b, 1 );
  s->add = b & HM_ADD;
  /* time */
  osc_write_str( o, "TBA?" );
  osc_read_str( o, "TBA:", &b, 1 );
  if ( ( b & HM_T_MASK ) > HM_T_MAX )
  {
    strcpy( o->error, "Illegal scaling of time axis."  );
    return 0;
  }
  s->Ts = HM_T_s_DIV[ b & HM_T_MASK ];
  osc_write_str( o, "HORMODE?" );
  osc_read_str( o, "HORMODE:", &b, 1 );
  s->XY = 0;
  if ( b & HM_XY ) s->XY = 1;
  s->ok = 1;
  fprintf( stderr, "b = %02x\n", b );
  if (! (b & HM_STORE) ) s->ok = 0;
  if ( (b & HM_TB_MASK) != HM_TBA ) s->ok = 0;
  return 1;
}


/**
 * Read data from cahannel n
 *
 * o	OSC struct
 * s	OSTATE struct
 * n	channel number
 * b	buffer of 2048 bytes
 */
void osc_get_chan( OSC * o, const OSTATE * s, int n, double * b )
{
    char str[13] = "RDWFM1:\x00\x00\x00\x08";
    int str_l = 11;
    unsigned char buf[2048];
    int i;

    if ( n == 2 )
	str[5] = '2';
    if ( strcmp( o->ver, "1.05" ) < 0 ) 
    {
	/* old versions need , and : */
	memcpy( &str[7], "\x00\x00,\x00\x08:", 6 );
	str_l += 2;
    }
    osc_write_strn( o, str, str_l );
    osc_read_strn( o, str, str_l, buf, 2048 );
    for ( i = 0; i < 2048; i++ )
	b[i] = ( buf[i] - 127 ) / 25.0 * s->Yv[n-1];
}


/**
 * Send two times break to scope.
 *
 * A break resets the serial communication of the scope.
 * Often (with a USB serial converter) it is useful to send one or two breaks
 * before to communication.
 *
 * c	   filename of serial port
 */
void osc_break( const char * c )
{
    int fd;
    char buf[256];

    fd = open( c, O_RDWR | O_NOCTTY );
    if ( fd == -1 ) 
    {
	strerror_r( errno, buf, 256 );
	fprintf( stderr, "Cannot send break: %s\n", buf );
    }
    else
    {
	fprintf( stderr, "--> break to %s\n", c );
	tcsendbreak(fd ,0);
	fprintf( stderr, "--> wait\n" );
	sleep(1);
    }
    close( fd );
}
