/**
 * HM acquire - software for reading out HAMEG digital storage
 * oscilloscopes.
 *         
 * Copyright (C) 2007 -- 2018 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@opppf.de
 * http://opppf.de/HMacquire/
 */


#ifndef OSC_API_H
#define OSC_API_H


/**
 * struct describing scope, reference for all functions
 */
typedef struct
{
  int fd;		/* file descriptor */
  char ver[5];		/* firmware version */
  char version[15];	/* firmware version, whole string */
  char error[256];	/* Error description as 0 terminated string, length 0
                           means no error */
} OSC;


/**
 * struct describing scope state
 */
typedef struct
{
  int Yon[2];	    /* 0 -> off, !0 -> on */
  int Ys[2];	    /* sign, +1 or -1 */
  int add;	    /* 0 = no add, !0 = add */
  int XY;	    /* 0 = no XY,  !0 = XY */
  double Yv[2];	    /* resolution in V/div */
  double Ts;	    /* resolution in s/div */
  char Yc[2];	    /* coupling: 'A' = ac, 'D' = dc */
  int ok;	    /* store mode */
} OSTATE;


/**
 * Open connection to scope.
 *
 * Set up serial port, initialise scope.
 * 
 * c       filename of serial port
 * return: OSC descriptor
 *         possible error in field error, error = "" means ok.
 */
OSC * osc_open( const char * c );


/**
 * Close Connection to scope.
 *
 * Switch back to manual mode.
 *
 * o	OSC struct
 */
void osc_close( OSC * o );

/**
 * Write 0 terminated string to scope.
 *
 * o	OSC struct
 * c	string to write
 * return: 1 = success; 0 = error
 */
int osc_write_str( OSC * o, const char * c );


/**
 * Write string with n bytes to scope.
 *
 * o	OSC struct
 * c	string to write
 * n	length of string
 * return: 1 = success; 0 = error
 */
int osc_write_strn( OSC * o, const char * c, int n );


/**
 * Read 0 terminated string from scope and store m further bytes to b.
 *
 * o	OSC struct
 * c	string to receive
 * b	destination for answer
 * m	bytes of answer
 * return: 1 = success; 0 = error
 */
int osc_read_str( OSC * o, const char * c, void * b, int m );


/**
 * Read  string with n bytes from scope and store m further bytes to b.
 *
 * o	OSC struct
 * c	string to receive
 * n	length of string
 * b	destination for answer
 * m	bytes of answer
 * return: 1 = success; 0 = error
 */
int osc_read_strn( OSC * o, const char * c, int n, void * b, int m );


/**
 * Read returncode R CR NL
 *
 * o	OSC struct
 * return: The R.
 */
char osc_get_returncode( OSC * o );


/**
 * Get scope state
 *
 * o	OSC struct
 * s	OSTATE struct
 * return: 1 = success; 0 = error
 */
int osc_get_state( OSC * o, OSTATE * s );


/**
 * Read data from cahannel n and scale data
 *
 * o	OSC struct
 * s	OSTATE struct
 * n	channel number
 * b	buffer, 2048 fields
 */
void osc_get_chan( OSC * o, const OSTATE * s, int n, double * b );


/**
 * Send two times break to scope.
 *
 * A break resets the serial communication of the scope.
 * Often (with a USB serial converter) it is useful to send one or two breaks
 * before to communication.
 *
 * c	   filename of serial port
 */
void osc_break( const char * c );

#endif
